@@ -30,7 +30,7 @@ use Storable;
 use Time::HiRes qw();
 use YAML::Syck;
 
-use version; our $VERSION = qv('0.0.5');
+use version; our $VERSION = qv('0.0.8');
 
 use constant MAX_INT => ~0>>1; # anything better?
 use constant DEFAULT_PROTOCOL => 1;
@@ -61,7 +61,6 @@ Reader/mirrorer:
     my $rf = File::Rsync::Mirror::Recentfile->new
       (
        filenameroot => "RECENT",
-       ignore_link_stat_errors => 1,
        interval => q(6h),
        localroot => "/home/ftp/pub/PAUSE/authors",
        remote_dir => "",
@@ -189,7 +188,7 @@ sub DESTROY {
     unless ($self->_current_tempfile_fh) {
         if (my $tempfile = $self->_current_tempfile) {
             if (-e $tempfile) {
-                unlink $tempfile; # may fail in global destruction
+                # unlink $tempfile; # may fail in global destruction
             }
         }
     }
@@ -215,6 +214,7 @@ BEGIN {
                   "_remember_last_uptodate_call",
                   "_remote_dir",
                   "_remoteroot",
+                  "_requires_fsck",
                   "_rfile",
                   "_rsync",
                   "__verified_tempdir",
@@ -265,8 +265,8 @@ Only relevant for slaves.
 
 If set to true, rsync errors are ignored that complain about link stat
 errors. These seem to happen only when there are files missing at the
-origin. In race conditions this can always happen, so it is
-recommended to set this value to true.
+origin. In race conditions this can always happen, so it defaults to
+true.
 
 =item is_slave
 
@@ -408,8 +408,9 @@ rate and as such are quite useless on their own.
 
 sub aggregate {
     my($self, %option) = @_;
+    my %seen_interval;
     my @aggs = sort { $a->{secs} <=> $b->{secs} }
-        grep { $_->{secs} >= $self->interval_secs }
+        grep { !$seen_interval{$_->{interval}}++ && $_->{secs} >= $self->interval_secs }
             map { { interval => $_, secs => $self->interval_secs($_)} }
                 $self->interval, @{$self->aggregator || []};
     $self->update;
@@ -602,6 +603,7 @@ sub get_remote_recentfile_as_tempfile {
     }
     my $gaveup = 0;
     my $retried = 0;
+    local($ENV{LANG}) = "C";
     while (!$self->rsync->exec(
                                src => $src,
                                dst => $dst,
@@ -706,9 +708,11 @@ sub get_remotefile {
              $path,
             );
     }
+    local($ENV{LANG}) = "C";
+    my $remoteroot = $self->remoteroot or die "Alert: missing remoteroot. Cannot continue";
     while (!$self->rsync->exec(
                                src => join("/",
-                                           $self->remoteroot,
+                                           $remoteroot,
                                            $path),
                                dst => $dst,
                               )) {
@@ -770,13 +774,15 @@ sub interval_secs {
 
 =head2 $obj->localroot ( $localroot )
 
-Get/set accessor. The local root of the tree.
+Get/set accessor. The local root of the tree. Guaranteed without
+trailing slash.
 
 =cut
 
 sub localroot {
     my ($self, $localroot) = @_;
     if (@_ >= 2) {
+        $localroot =~ s|/$||;
         $self->_localroot($localroot);
         $self->_rfile(undef);
     }
@@ -819,12 +825,35 @@ sub lock {
     # XXX need a way to allow breaking the lock
     my $start = time;
     my $locktimeout = $self->locktimeout || 600;
-    while (not mkdir "$rfile.lock") {
+    my %have_warned;
+    my $lockdir = "$rfile.lock";
+    my $procfile = "$lockdir/process";
+ GETLOCK: while (not mkdir $lockdir) {
+        if (open my $fh, "<", $procfile) {
+            chomp(my $process = <$fh>);
+            if (0) {
+            } elsif ($process !~ /^\d+$/) {
+                warn "Warning: unknown process holds a lock in '$lockdir', waiting..." unless $have_warned{unknown}++;
+            } elsif ($$ == $process) {
+                last GETLOCK;
+            } elsif (kill 0, $process) {
+                warn "Warning: process $process holds a lock in '$lockdir', waiting..." unless $have_warned{$process}++;
+            } else {
+                warn "Warning: breaking lock held by process $process";
+                sleep 1;
+                last GETLOCK;
+            }
+        } else {
+            warn "Warning: unknown process holds a lock in '$lockdir', waiting..." unless $have_warned{unknown}++;
+        }
         Time::HiRes::sleep 0.01;
         if (time - $start > $locktimeout) {
             die "Could not acquire lockdirectory '$rfile.lock': $!";
         }
-    }
+    } # GETLOCK
+    open my $fh, ">", $procfile or die "Could not open >$procfile\: $!";
+    print $fh $$, "\n";
+    close $fh or die "Could not close: $!";
     $self->_is_locked (1);
 }
 
@@ -846,7 +875,6 @@ sub merge {
     $self->_merge_sanitycheck ( $other );
     $other->lock;
     my $other_recent = $other->recent_events || [];
-    # $DB::single++ if $other->interval_secs eq "2" and grep {$_->{epoch} eq "999.999"} @$other_recent;
     $self->lock;
     $self->_merge_locked ( $other, $other_recent );
     $self->unlock;
@@ -955,12 +983,14 @@ sub _merge_something_done {
 sub _merge_sanitycheck {
     my($self, $other) = @_;
     if ($self->interval_secs <= $other->interval_secs) {
-        die sprintf
-            (
-             "Alert: illegal merge operation of a bigger interval[%d] into a smaller[%d]",
-             $self->interval_secs,
-             $other->interval_secs,
-            );
+        require Carp;
+        Carp::confess
+                (sprintf
+                 (
+                  "Alert: illegal merge operation of a bigger interval[%d] into a smaller[%d]",
+                  $self->interval_secs,
+                  $other->interval_secs,
+                 ));
     }
 }
 
@@ -1043,12 +1073,11 @@ sub meta_data {
 =head2 $success = $obj->mirror ( %options )
 
 Mirrors the files in this I<recentfile> as reported by
-C<recent_events>. Options named C<after>, C<before>, C<max>, and
-C<skip-deletes> are passed through to the C<recent_events> call. The
-boolean option C<piecemeal>, if true, causes C<mirror> to only rsync
-C<max_files_per_connection> and keep track of the rsynced files so
-that future calls will rsync different files until all files are
-brought to sync.
+C<recent_events>. Options named C<after>, C<before>, C<max> are passed
+through to the C<recent_events> call. The boolean option C<piecemeal>,
+if true, causes C<mirror> to only rsync C<max_files_per_connection>
+and keep track of the rsynced files so that future calls will rsync
+different files until all files are brought to sync.
 
 =cut
 
@@ -1056,7 +1085,10 @@ sub mirror {
     my($self, %options) = @_;
     my $trecentfile = $self->get_remote_recentfile_as_tempfile();
     $self->_use_tempfile (1);
-    my %passthrough = map { ($_ => $options{$_}) } qw(before after max skip-deletes);
+    # skip-deletes is inadequat for passthrough within mirror. We
+    # would never reach uptodateness when a delete were on a
+    # borderline
+    my %passthrough = map { ($_ => $options{$_}) } qw(before after max);
     my ($recent_events) = $self->recent_events(%passthrough);
     my(@error, @dlcollector); # download-collector: array containing paths we need
     my $first_item = 0;
@@ -1104,7 +1136,7 @@ sub mirror {
     # once we've gone to the end we consider ourselves free of obligations
     $self->unseed;
     $self->_mirror_unhide_tempfile ($trecentfile);
-    $self->_mirror_perform_delayed_ops;
+    $self->_mirror_perform_delayed_ops(\%options);
     return !@error;
 }
 
@@ -1268,21 +1300,45 @@ sub _mirror_unhide_tempfile {
 }
 
 sub _mirror_perform_delayed_ops {
-    my($self) = @_;
+    my($self,$options) = @_;
     my $delayed = $self->delayed_operations;
     for my $dst (keys %{$delayed->{unlink}}) {
         unless (unlink $dst) {
             require Carp;
-            Carp::cluck ( "Warning: Error while unlinking '$dst': $!" );
+            Carp::cluck ( "Warning: Error while unlinking '$dst': $!" ) if $options->{verbose};
+        }
+        if ($self->verbose) {
+            my $doing = "Del";
+            my $LFH = $self->_logfilehandle;
+            printf $LFH
+                (
+                 "%-4s %d (%s) %s DONE\n",
+                 $doing,
+                 time,
+                 $self->interval,
+                 $dst,
+                );
+            delete $delayed->{unlink}{$dst};
         }
-        delete $delayed->{unlink}{$dst};
     }
     for my $dst (sort {length($b) <=> length($a)} keys %{$delayed->{rmdir}}) {
         unless (rmdir $dst) {
             require Carp;
-            Carp::cluck ( "Warning: Error on rmdir '$dst': $!" );
+            Carp::cluck ( "Warning: Error on rmdir '$dst': $!" ) if $options->{verbose};
+        }
+        if ($self->verbose) {
+            my $doing = "Del";
+            my $LFH = $self->_logfilehandle;
+            printf $LFH
+                (
+                 "%-4s %d (%s) %s DONE\n",
+                 $doing,
+                 time,
+                 $self->interval,
+                 $dst,
+                );
+            delete $delayed->{rmdir}{$dst};
         }
-        delete $delayed->{rmdir}{$dst};
     }
 }
 
@@ -1321,6 +1377,7 @@ sub mirror_path {
         $fh->unlink_on_destroy(1);
         my $gaveup = 0;
         my $retried = 0;
+        local($ENV{LANG}) = "C";
         while (!$self->rsync->exec
                (
                 src => join("/",
@@ -1330,7 +1387,7 @@ sub mirror_path {
                 'files-from' => $fh->filename,
                )) {
             my(@err) = $self->rsync->err;
-            if ($self->ignore_link_stat_errors && "@err" =~ m{^ rsync: \s link_stat }x ) {
+            if ($self->_my_ignore_link_stat_errors && "@err" =~ m{^ rsync: \s link_stat }x ) {
                 if ($self->verbose) {
                     my $LFH = $self->_logfilehandle;
                     print $LFH "Info: ignoring link_stat error '@err'";
@@ -1339,10 +1396,12 @@ sub mirror_path {
             }
             $self->register_rsync_error (@err);
             if (++$retried >= 3) {
-                warn "XXX giving up.";
+                my $batchsize = @$path;
+                warn "The number of rsync retries now reached 3 within a batch of size $batchsize. Error was '@err'. Giving up now, will retry later, ";
                 $gaveup = 1;
                 last;
             }
+            sleep 1;
         }
         unless ($gaveup) {
             $self->un_register_rsync_error ();
@@ -1350,6 +1409,7 @@ sub mirror_path {
     } else {
         my $dst = $self->local_path($path);
         mkpath dirname $dst;
+        local($ENV{LANG}) = "C";
         while (!$self->rsync->exec
                (
                 src => join("/",
@@ -1359,7 +1419,7 @@ sub mirror_path {
                 dst => $dst,
                 )) {
             my(@err) = $self->rsync->err;
-            if ($self->ignore_link_stat_errors && "@err" =~ m{^ rsync: \s link_stat }x ) {
+            if ($self->_my_ignore_link_stat_errors && "@err" =~ m{^ rsync: \s link_stat }x ) {
                 if ($self->verbose) {
                     my $LFH = $self->_logfilehandle;
                     print $LFH "Info: ignoring link_stat error '@err'";
@@ -1373,12 +1433,20 @@ sub mirror_path {
     return 1;
 }
 
+sub _my_ignore_link_stat_errors {
+    my($self) = @_;
+    my $x = $self->ignore_link_stat_errors;
+    $x = 1 unless defined $x;
+    return $x;
+}
+
 sub _my_current_rfile {
     my($self) = @_;
     my $rfile;
     if ($self->_use_tempfile) {
         $rfile = $self->_current_tempfile;
-    } else {
+    }
+    unless ($rfile && -s $rfile) {
         $rfile = $self->rfile;
     }
     return $rfile;
@@ -1426,8 +1494,8 @@ timestamp are returned.
 If C<$options{before}> is specified, only file events before this
 timestamp are returned.
 
-If C<$options{max}> is specified only a maximum of this many events is
-returned.
+If C<$options{max}> is specified only a maximum of this many most
+recent events is returned.
 
 If C<$options{'skip-deletes'}> is specified, no files-to-be-deleted
 will be returned.
@@ -1550,7 +1618,10 @@ sub _recent_events_protocol_x {
     my $meth = sprintf "read_recent_%d", $data->{meta}{protocol};
     # we may be reading meta for the first time
     while (my($k,$v) = each %{$data->{meta}}) {
-        next if $k ne lc $k; # "Producers"
+        if ($k ne lc $k){ # "Producers"
+            $self->{ORIG}{$k} = $v;
+            next;
+        }
         next if defined $self->$k;
         $self->$k($v);
     }
@@ -1828,6 +1899,7 @@ sub _sparse_clone {
                   _use_tempfile
                   aggregator
                   filenameroot
+                  ignore_link_stat_errors
                   is_slave
                   max_files_per_connection
                   protocol
@@ -1871,7 +1943,8 @@ sub unlock {
     my($self) = @_;
     return unless $self->_is_locked;
     my $rfile = $self->rfile;
-    rmdir "$rfile.lock";
+    unlink "$rfile.lock/process" or warn "Could not unlink lockfile '$rfile.lock/process': $!";
+    rmdir "$rfile.lock" or warn "Could not rmdir lockdir '$rfile.lock': $!";;
     $self->_is_locked (0);
 }
 
@@ -1935,43 +2008,78 @@ sub update {
         die "update called without path argument" unless defined $path;
         die "update called without type argument" unless defined $type;
         die "update called with illegal type argument: $type" unless $type =~ /(new|delete)/;
-        # since we have keep_delete_objects_forever we must let them inject delete objects too:
-        #die "update called with \$type=$type and \$dirty_epoch=$dirty_epoch; ".
-        #    "dirty_epoch only allowed with type=new" if defined $dirty_epoch and $type ne "new";
-        my $canonmeth = $self->canonize;
-        unless ($canonmeth) {
-            $canonmeth = "naive_path_normalize";
-        }
-        $path = $self->$canonmeth($path);
     }
-    my $lrd = $self->localroot;
     $self->lock;
-    # you must calculate the time after having locked, of course
-    my $now = Time::HiRes::time;
-    my $interval = $self->interval;
-    my $secs = $self->interval_secs();
-    my $recent = $self->recent_events;
+    my $ctx = $self->_locked_batch_update([{path=>$path,type=>$type,epoch=>$dirty_epoch}]);
+    $self->write_recent($ctx->{recent}) if $ctx->{something_done};
+    $self->_assert_symlink;
+    $self->unlock;
+}
 
-    my $epoch;
-    if (defined $dirty_epoch && _bigfloatgt($now,$dirty_epoch)) {
-        $epoch = $dirty_epoch;
-    } else {
-        $epoch = $self->_epoch_monotonically_increasing($now,$recent);
-    }
+=head2 $obj->batch_update($batch)
 
-    $recent ||= [];
-    my $oldest_allowed = 0;
-    my $merged = $self->merged;
-    if ($merged->{epoch}) {
-        my $virtualnow = _bigfloatmax($now,$epoch);
-        # for the lower bound I think we need no big math, we calc already
-        $oldest_allowed = min($virtualnow - $secs, $merged->{epoch}, $epoch);
-    } else {
-        # as long as we are not merged at all, no limits!
-    }
+Like update but for many files. $batch is an arrayref containing
+hashrefs with the structure
+
+  {
+    path => $path,
+    type => $type,
+    epoch => $epoch,
+  }
+
+
+
+=cut
+sub batch_update {
+    my($self,$batch) = @_;
+    $self->lock;
+    my $ctx = $self->_locked_batch_update($batch);
+    $self->write_recent($ctx->{recent}) if $ctx->{something_done};
+    $self->_assert_symlink;
+    $self->unlock;
+}
+sub _locked_batch_update {
+    my($self,$batch) = @_;
     my $something_done = 0;
- TRUNCATE: while (@$recent) {
-        # $DB::single++ unless defined $oldest_allowed;
+    my $recent = $self->recent_events;
+    unless ($recent->[0]) {
+        # obstetrics
+        $something_done = 1;
+    }
+    my %paths_in_recent = map { $_->{path} => undef } @$recent;
+    my $interval = $self->interval;
+    my $canonmeth = $self->canonize;
+    unless ($canonmeth) {
+        $canonmeth = "naive_path_normalize";
+    }
+    my $oldest_allowed = 0;
+    my $setting_new_dirty_mark = 0;
+    my $console;
+    if ($self->verbose && @$batch > 1) {
+        eval {require Time::Progress};
+        warn "dollarat[$@]" if $@;
+        $| = 1;
+        $console = new Time::Progress;
+        $console->attr( min => 1, max => scalar @$batch );
+        print "\n";
+    }
+    my $i = 0;
+    my $memo_splicepos;
+ ITEM: for my $item (sort {($b->{epoch}||0) <=> ($a->{epoch}||0)} @$batch) {
+        $i++;
+        print $console->report( "\rdone %p elapsed: %L (%l sec), ETA %E (%e sec)", $i ) if $console and not $i % 50;
+        my $ctx = $self->_update_batch_item($item,$canonmeth,$recent,$setting_new_dirty_mark,$oldest_allowed,$something_done,\%paths_in_recent,$memo_splicepos);
+        $something_done = $ctx->{something_done};
+        $oldest_allowed = $ctx->{oldest_allowed};
+        $setting_new_dirty_mark = $ctx->{setting_new_dirty_mark};
+        $recent = $ctx->{recent};
+        $memo_splicepos = $ctx->{memo_splicepos};
+    }
+    print "\n" if $console;
+    if ($setting_new_dirty_mark) {
+        $oldest_allowed = 0;
+    }
+TRUNCATE: while (@$recent) {
         if (_bigfloatlt($recent->[-1]{epoch}, $oldest_allowed)) {
             pop @$recent;
             $something_done = 1;
@@ -1979,12 +2087,40 @@ sub update {
             last TRUNCATE;
         }
     }
+    return {something_done=>$something_done,recent=>$recent};
+}
+sub _update_batch_item {
+    my($self,$item,$canonmeth,$recent,$setting_new_dirty_mark,$oldest_allowed,$something_done,$paths_in_recent,$memo_splicepos) = @_;
+    my($path,$type,$dirty_epoch) = @{$item}{qw(path type epoch)};
+    if (defined $path or defined $type or defined $dirty_epoch) {
+        $path = $self->$canonmeth($path);
+    }
+    # you must calculate the time after having locked, of course
+    my $now = Time::HiRes::time;
+
+    my $epoch;
+    if (defined $dirty_epoch && _bigfloatgt($now,$dirty_epoch)) {
+        $epoch = $dirty_epoch;
+    } else {
+        $epoch = $self->_epoch_monotonically_increasing($now,$recent);
+    }
+    $recent ||= [];
+    my $merged = $self->merged;
+    if ($merged->{epoch} && !$setting_new_dirty_mark) {
+        my $virtualnow = _bigfloatmax($now,$epoch);
+        # for the lower bound I think we need no big math, we calc already
+        my $secs = $self->interval_secs();
+        $oldest_allowed = min($virtualnow - $secs, $merged->{epoch}, $epoch);
+        } else {
+            # as long as we are not merged at all, no limits!
+        }
+    my $lrd = $self->localroot;
     if (defined $path && $path =~ s|^\Q$lrd\E||) {
         $path =~ s|^/||;
         my $splicepos;
         # remove the older duplicates of this $path, irrespective of $type:
         if (defined $dirty_epoch) {
-            my $ctx = $self->_update_with_dirty_epoch($path,$recent,$epoch);
+            my $ctx = $self->_update_with_dirty_epoch($path,$recent,$epoch,$paths_in_recent,$memo_splicepos);
             $recent    = $ctx->{recent};
             $splicepos = $ctx->{splicepos};
             $epoch     = $ctx->{epoch};
@@ -1994,7 +2130,7 @@ sub update {
                 $new_dm = $epoch;
             }
             $self->dirtymark($new_dm);
-            my $merged = $self->merged;
+            $setting_new_dirty_mark = 1;
             if (not defined $merged->{epoch} or _bigfloatlt($epoch,$merged->{epoch})) {
                 $self->merged(+{});
             }
@@ -2004,20 +2140,25 @@ sub update {
         }
         if (defined $splicepos) {
             splice @$recent, $splicepos, 0, { epoch => $epoch, path => $path, type => $type };
+            $paths_in_recent->{$path} = undef;
         }
+        $memo_splicepos = $splicepos;
         $something_done = 1;
     }
-
-    $self->write_recent($recent) if $something_done;
-    $self->_assert_symlink;
-    $self->unlock;
+    return
+        {
+         something_done => $something_done,
+         oldest_allowed => $oldest_allowed,
+         setting_new_dirty_mark => $setting_new_dirty_mark,
+         recent => $recent,
+         memo_splicepos => $memo_splicepos,
+        }
 }
-
 sub _update_with_dirty_epoch {
-    my($self,$path,$recent,$epoch) = @_;
+    my($self,$path,$recent,$epoch,$paths_in_recent,$memo_splicepos) = @_;
     my $splicepos;
     my $new_recent = [];
-    if (grep { $_->{path} ne $path } @$recent) {
+    if (exists $paths_in_recent->{$path}) {
         my $cancel = 0;
     KNOWN_EVENT: for my $i (0..$#$recent) {
             if ($recent->[$i]{path} eq $path) {
@@ -2037,7 +2178,13 @@ sub _update_with_dirty_epoch {
     } elsif (_bigfloatlt($epoch,$recent->[-1]{epoch})) {
         $splicepos = @$recent;
     } else {
-    RECENT: for my $i (0..$#$recent) {
+        my $startingpoint;
+        if (_bigfloatgt($memo_splicepos<=$#$recent && $epoch, $recent->[$memo_splicepos]{epoch})) {
+            $startingpoint = 0;
+        } else {
+            $startingpoint = $memo_splicepos;
+        }
+    RECENT: for my $i ($startingpoint..$#$recent) {
             my $ev = $recent->[$i];
             if ($epoch eq $recent->[$i]{epoch}) {
                 $epoch = _increase_a_bit($epoch, $i ? $recent->[$i-1]{epoch} : undef);
@@ -2111,15 +2258,22 @@ sub uptodate {
         my $minmax = $self->minmax;
         if (exists $minmax->{mtime}) {
             my $rfile = $self->_my_current_rfile;
-            my @stat = stat $rfile or die "Could not stat '$rfile': $!";
-            my $mtime = $stat[9];
-            if (defined $mtime && defined $minmax->{mtime} && $mtime > $minmax->{mtime}) {
-                $why = "mtime[$mtime] of rfile[$rfile] > minmax/mtime[$minmax->{mtime}], so we are not uptodate";
-                $uptodate = 0;
+            my @stat = stat $rfile;
+            if (@stat) {
+                my $mtime = $stat[9];
+                if (defined $mtime && defined $minmax->{mtime} && $mtime > $minmax->{mtime}) {
+                    $why = "mtime[$mtime] of rfile[$rfile] > minmax/mtime[$minmax->{mtime}], so we are not uptodate";
+                    $uptodate = 0;
+                } else {
+                    my $covered = $self->done->covered(@$minmax{qw(max min)});
+                    $why = sprintf "minmax covered[%s], so we return that", defined $covered ? $covered : "UNDEF";
+                    $uptodate = $covered;
+                }
             } else {
-                my $covered = $self->done->covered(@$minmax{qw(max min)});
-                $why = sprintf "minmax covered[%s], so we return that", defined $covered ? $covered : "UNDEF";
-                $uptodate = $covered;
+                require Carp;
+                $why = "Could not stat '$rfile': $!";
+                Carp::cluck($why);
+                $uptodate = 0;
             }
         }
     }
@@ -2167,10 +2321,10 @@ sub write_recent {
     }
     my $minmax = $self->minmax;
     if (!defined $minmax->{max} || _bigfloatlt($minmax->{max},$recent->[0]{epoch})) {
-        $minmax->{max} = $recent->[0]{epoch};
+        $minmax->{max} = @$recent && exists $recent->[0]{epoch} ? $recent->[0]{epoch} : undef;
     }
     if (!defined $minmax->{min} || _bigfloatlt($minmax->{min},$recent->[-1]{epoch})) {
-        $minmax->{min} = $recent->[-1]{epoch};
+        $minmax->{min} = @$recent && exists $recent->[-1]{epoch} ? $recent->[-1]{epoch} : undef;
     }
     $self->minmax($minmax);
     my $meth = sprintf "write_%d", $self->protocol;
